@@ -5,6 +5,7 @@ Build all of the packages in a given directory.
 """
 
 import argparse
+import copy
 import dataclasses
 import hashlib
 import json
@@ -12,6 +13,7 @@ import os
 import shutil
 import subprocess
 import sys
+from collections import defaultdict
 from collections.abc import Iterable
 from functools import total_ordering
 from graphlib import TopologicalSorter
@@ -23,8 +25,8 @@ from typing import Any
 
 from . import common
 from .buildpkg import needs_rebuild
-from .common import UNVENDORED_STDLIB_MODULES, find_matching_wheels
-from .io import parse_package_config
+from .common import find_matching_wheels, find_missing_executables
+from .io import MetaConfig, _BuildSpecTypes
 
 
 class BuildError(Exception):
@@ -40,13 +42,14 @@ class BasePackage:
     name: str
     version: str
     disabled: bool
-    cpython_dynlib: bool
-    meta: dict[str, Any]
-    library: bool
-    shared_library: bool
-    dependencies: list[str]
-    unbuilt_dependencies: set[str]
-    dependents: set[str]
+    meta: MetaConfig
+    package_type: _BuildSpecTypes
+    run_dependencies: list[str]
+    host_dependencies: list[str]
+    executables_required: list[str]
+    dependencies: set[str]  # run + host dependencies
+    unbuilt_host_dependencies: set[str]
+    host_dependents: set[str]
     unvendored_tests: Path | None = None
     file_name: str | None = None
     install_dir: str = "site"
@@ -55,55 +58,27 @@ class BasePackage:
     # We use this in the priority queue, which pops off the smallest element.
     # So we want the smallest element to have the largest number of dependents
     def __lt__(self, other: Any) -> bool:
-        return len(self.dependents) > len(other.dependents)
+        return len(self.host_dependents) > len(other.host_dependents)
 
     def __eq__(self, other: Any) -> bool:
-        return len(self.dependents) == len(other.dependents)
+        return len(self.host_dependents) == len(other.host_dependents)
 
     def __repr__(self) -> str:
         return f"{type(self).__name__}({self.name})"
 
     def needs_rebuild(self) -> bool:
-        return needs_rebuild(
-            self.pkgdir, self.pkgdir / "build", self.meta.get("source", {})
-        )
+        return needs_rebuild(self.pkgdir, self.pkgdir / "build", self.meta.source)
 
-    def build(self, outputdir: Path, args: Any) -> None:
+    def build(self, args: Any) -> None:
         raise NotImplementedError()
 
-    def wheel_path(self) -> Path:
+    def dist_artifact_path(self) -> Path:
         raise NotImplementedError()
 
     def tests_path(self) -> Path | None:
         return None
 
 
-@dataclasses.dataclass
-class StdLibPackage(BasePackage):
-    def __init__(self, pkgdir: Path):
-        self.pkgdir = pkgdir
-        self.meta = {}
-        self.name = pkgdir.stem
-        self.version = "1.0"
-        self.disabled = False
-        self.library = False
-        self.shared_library = False
-        self.dependencies = []
-        self.unbuilt_dependencies = set()
-        self.dependents = set()
-        self.install_dir = "lib"
-
-    def build(self, outputdir: Path, args: Any) -> None:
-        # All build / packaging steps are already done in the main Makefile
-        return
-
-    def wheel_path(self) -> Path:
-        raise RuntimeError("StdLibPackage has no wheel")
-
-    def tests_path(self) -> Path | None:
-        return None
-
-
 @dataclasses.dataclass
 class Package(BasePackage):
     def __init__(self, pkgdir: Path):
@@ -113,31 +88,34 @@ class Package(BasePackage):
         if not pkgpath.is_file():
             raise ValueError(f"Directory {pkgdir} does not contain meta.yaml")
 
-        self.meta = parse_package_config(pkgpath)
-        self.name = self.meta["package"]["name"]
-        self.version = self.meta["package"]["version"]
-        self.disabled = self.meta["package"].get("_disabled", False)
-        self.cpython_dynlib = self.meta["package"].get("_cpython_dynlib", False)
-        self.meta["build"] = self.meta.get("build", {})
-        self.meta["requirements"] = self.meta.get("requirements", {})
+        self.meta = MetaConfig.from_yaml(pkgpath)
+        self.name = self.meta.package.name
+        self.version = self.meta.package.version
+        self.disabled = self.meta.package.disabled
+        self.package_type = self.meta.build.package_type
 
-        self.library = self.meta["build"].get("library", False)
-        self.shared_library = self.meta["build"].get("sharedlibrary", False)
+        assert self.name == pkgdir.name, f"{self.name} != {pkgdir.name}"
 
-        assert self.name == pkgdir.stem
+        self.run_dependencies = self.meta.requirements.run
+        self.host_dependencies = self.meta.requirements.host
+        self.executables_required = self.meta.requirements.executable
+        self.dependencies = set(self.run_dependencies + self.host_dependencies)
+        self.unbuilt_host_dependencies = set(self.host_dependencies)
+        self.host_dependents = set()
 
-        self.dependencies = self.meta["requirements"].get("run", [])
-        self.unbuilt_dependencies = set(self.dependencies)
-        self.dependents = set()
-
-    def wheel_path(self) -> Path:
+    def dist_artifact_path(self) -> Path:
         dist_dir = self.pkgdir / "dist"
-        wheels = list(find_matching_wheels(dist_dir.glob("*.whl")))
-        if len(wheels) != 1:
+        if self.package_type in ("shared_library", "cpython_module"):
+            candidates = list(dist_dir.glob("*.zip"))
+        else:
+            candidates = list(find_matching_wheels(dist_dir.glob("*.whl")))
+
+        if len(candidates) != 1:
             raise RuntimeError(
-                f"Unexpected number of wheels {len(wheels)} when building {self.name}"
+                f"Unexpected number of wheels/archives {len(candidates)} when building {self.name}"
             )
-        return wheels[0]
+
+        return candidates[0]
 
     def tests_path(self) -> Path | None:
         tests = list((self.pkgdir / "dist").glob("*-tests.tar"))
@@ -146,7 +124,7 @@ class Package(BasePackage):
             return tests[0]
         return None
 
-    def build(self, outputdir: Path, args: Any) -> None:
+    def build(self, args: Any) -> None:
 
         p = subprocess.run(
             [
@@ -155,16 +133,11 @@ class Package(BasePackage):
                 "pyodide_build",
                 "buildpkg",
                 str(self.pkgdir / "meta.yaml"),
-                "--cflags",
-                args.cflags,
-                "--cxxflags",
-                args.cxxflags,
-                "--ldflags",
-                args.ldflags,
-                "--target-install-dir",
-                args.target_install_dir,
-                "--host-install-dir",
-                args.host_install_dir,
+                f"--cflags={args.cflags}",
+                f"--cxxflags={args.cxxflags}",
+                f"--ldflags={args.ldflags}",
+                f"--target-install-dir={args.target_install_dir}",
+                f"--host-install-dir={args.host_install_dir}",
                 # Either this package has been updated and this doesn't
                 # matter, or this package is dependent on a package that has
                 # been updated and should be rebuilt even though its own
@@ -193,13 +166,15 @@ class Package(BasePackage):
             print("ERROR: cancelling buildall")
             raise BuildError(p.returncode)
 
-        if self.library:
-            return
-        if self.shared_library:
-            file_path = Path(self.pkgdir / f"{self.name}-{self.version}.zip")
-            shutil.copy(file_path, outputdir)
-            file_path.unlink()
-            return
+
+def validate_dependencies(pkg_map: dict[str, BasePackage]) -> None:
+    for pkg_name, pkg in pkg_map.items():
+        for runtime_dep_name in pkg.run_dependencies:
+            runtime_dep = pkg_map[runtime_dep_name]
+            if runtime_dep.package_type == "static_library":
+                raise ValueError(
+                    f"{pkg_name} has an invalid dependency: {runtime_dep_name}. Static libraries must be a host dependency."
+                )
 
 
 def generate_dependency_graph(
@@ -258,16 +233,15 @@ def generate_dependency_graph(
     while packages:
         pkgname = packages.pop()
 
-        if pkgname in UNVENDORED_STDLIB_MODULES:
-            pkg = StdLibPackage(packages_dir / pkgname)
-        else:
-            pkg = Package(packages_dir / pkgname)
+        pkg = Package(packages_dir / pkgname)
         pkg_map[pkgname] = pkg
         graph[pkgname] = pkg.dependencies
         for dep in pkg.dependencies:
             if pkg_map.get(dep) is None:
                 packages.add(dep)
 
+    validate_dependencies(pkg_map)
+
     # Traverse in build order (dependencies first then dependents)
     # Mark a package as disabled if they've either been explicitly disabled
     # or if any of its transitive dependencies were marked disabled.
@@ -298,8 +272,22 @@ def generate_dependency_graph(
             continue
 
         requested.update(pkg.dependencies)
-        for dep in pkg.dependencies:
-            pkg_map[dep].dependents.add(pkg.name)
+        for dep in pkg.host_dependencies:
+            pkg_map[dep].host_dependents.add(pkg.name)
+
+    # Check executables required to build packages
+    missing_executables = defaultdict(list)
+    for name in requested:
+        pkg = pkg_map[name]
+        for exe in find_missing_executables(pkg.executables_required):
+            missing_executables[exe].append(name)
+
+    if missing_executables:
+        error_msg = "The following executables are missing in the host system:\n"
+        for executable, pkgs in missing_executables.items():
+            error_msg += f"- {executable} (required by: {', '.join(pkgs)})\n"
+
+        raise RuntimeError(error_msg)
 
     return {name: pkg_map[name] for name in requested}
 
@@ -352,12 +340,10 @@ def mark_package_needs_build(
     Helper for generate_needs_build_set. Modifies needs_build in place.
     Recursively add pkg and all of its dependencies to needs_build.
     """
-    if isinstance(pkg, StdLibPackage):
-        return
     if pkg.name in needs_build:
         return
     needs_build.add(pkg.name)
-    for dep in pkg.dependents:
+    for dep in pkg.host_dependents:
         mark_package_needs_build(pkg_map, pkg_map[dep], needs_build)
 
 
@@ -378,9 +364,7 @@ def generate_needs_build_set(pkg_map: dict[str, BasePackage]) -> set[str]:
     return needs_build
 
 
-def build_from_graph(
-    pkg_map: dict[str, BasePackage], outputdir: Path, args: argparse.Namespace
-) -> None:
+def build_from_graph(pkg_map: dict[str, BasePackage], args: argparse.Namespace) -> None:
     """
     This builds packages in pkg_map in parallel, building at most args.n_jobs
     packages at once.
@@ -414,7 +398,7 @@ def build_from_graph(
     # Remove the packages we've already built from the dependency sets of
     # the remaining ones
     for pkg_name in needs_build:
-        pkg_map[pkg_name].unbuilt_dependencies.difference_update(already_built)
+        pkg_map[pkg_name].unbuilt_host_dependencies.difference_update(already_built)
 
     if already_built:
         print(
@@ -428,7 +412,7 @@ def build_from_graph(
     t0 = perf_counter()
     for pkg_name in needs_build:
         pkg = pkg_map[pkg_name]
-        if len(pkg.unbuilt_dependencies) == 0:
+        if len(pkg.unbuilt_host_dependencies) == 0:
             build_queue.put((job_priority(pkg), pkg))
 
     built_queue: Queue[BasePackage | Exception] = Queue()
@@ -450,7 +434,7 @@ def build_from_graph(
             t0 = perf_counter()
             success = True
             try:
-                pkg.build(outputdir, args)
+                pkg.build(args)
             except Exception as e:
                 built_queue.put(e)
                 success = False
@@ -484,10 +468,10 @@ def build_from_graph(
 
         num_built += 1
 
-        for _dependent in pkg.dependents:
+        for _dependent in pkg.host_dependents:
             dependent = pkg_map[_dependent]
-            dependent.unbuilt_dependencies.remove(pkg.name)
-            if len(dependent.unbuilt_dependencies) == 0:
+            dependent.unbuilt_host_dependencies.remove(pkg.name)
+            if len(dependent.unbuilt_host_dependencies) == 0:
                 build_queue.put((job_priority(dependent), dependent))
 
     print(
@@ -507,10 +491,9 @@ def _generate_package_hash(full_path: Path) -> str:
 def generate_packagedata(
     output_dir: Path, pkg_map: dict[str, BasePackage]
 ) -> dict[str, Any]:
-    libraries = [pkg.name for pkg in pkg_map.values() if pkg.library]
     packages: dict[str, Any] = {}
     for name, pkg in pkg_map.items():
-        if not pkg.file_name:
+        if not pkg.file_name or pkg.package_type == "static_library":
             continue
         if not Path(output_dir, pkg.file_name).exists():
             continue
@@ -521,14 +504,19 @@ def generate_packagedata(
             "install_dir": pkg.install_dir,
             "sha256": _generate_package_hash(Path(output_dir, pkg.file_name)),
         }
-        if pkg.shared_library:
-            pkg_entry["shared_library"] = True
-            pkg_entry["install_dir"] = "lib" if pkg.cpython_dynlib else "dynlib"
 
-        pkg_entry["depends"] = [
-            x.lower() for x in pkg.dependencies if x not in libraries
-        ]
-        pkg_entry["imports"] = pkg.meta.get("test", {}).get("imports", [name])
+        pkg_type = pkg.package_type
+        if pkg_type in ("shared_library", "cpython_module"):
+            # We handle cpython modules as shared libraries
+            pkg_entry["shared_library"] = True
+            pkg_entry["install_dir"] = (
+                "stdlib" if pkg_type == "cpython_module" else "dynlib"
+            )
+
+        pkg_entry["depends"] = [x.lower() for x in pkg.run_dependencies]
+        pkg_entry["imports"] = (
+            pkg.meta.package.top_level if pkg.meta.package.top_level else [name]
+        )
 
         packages[name.lower()] = pkg_entry
 
@@ -549,11 +537,6 @@ def generate_packagedata(
             }
             packages[name.lower() + "-tests"] = pkg_entry
 
-    # Workaround for circular dependency between soupsieve and beautifulsoup4
-    # TODO: FIXME!!
-    if "soupsieve" in packages:
-        packages["soupsieve"]["depends"].append("beautifulsoup4")
-
     # sort packages by name
     packages = dict(sorted(packages.items()))
     return packages
@@ -585,10 +568,10 @@ def copy_packages_to_dist_dir(
     packages: Iterable[BasePackage], output_dir: Path
 ) -> None:
     for pkg in packages:
-        try:
-            shutil.copy(pkg.wheel_path(), output_dir)
-        except RuntimeError:
-            pass
+        if pkg.package_type == "static_library":
+            continue
+
+        shutil.copy(pkg.dist_artifact_path(), output_dir)
 
         test_path = pkg.tests_path()
         if test_path:
@@ -596,28 +579,39 @@ def copy_packages_to_dist_dir(
 
 
 def build_packages(
-    packages_dir: Path, output_dir: Path, args: argparse.Namespace
-) -> None:
+    packages_dir: Path, args: argparse.Namespace
+) -> dict[str, BasePackage]:
     packages = common._parse_package_subset(args.only)
-
     pkg_map = generate_dependency_graph(packages_dir, packages)
 
-    build_from_graph(pkg_map, output_dir, args)
+    build_from_graph(pkg_map, args)
     for pkg in pkg_map.values():
-        if pkg.library:
-            continue
-        if isinstance(pkg, StdLibPackage):
-            pkg.file_name = pkg.name + ".tar"
-            continue
-        if pkg.needs_rebuild():
-            continue
-        if pkg.shared_library:
-            pkg.file_name = f"{pkg.name}-{pkg.version}.zip"
-            continue
         assert isinstance(pkg, Package)
-        pkg.file_name = pkg.wheel_path().name
+
+        if pkg.package_type == "static_library":
+            continue
+
+        pkg.file_name = pkg.dist_artifact_path().name
         pkg.unvendored_tests = pkg.tests_path()
 
+    return pkg_map
+
+
+def install_packages(pkg_map: dict[str, BasePackage], output_dir: Path) -> None:
+    """
+    Install packages into the output directory.
+    - copies build artifacts (wheel, zip, ...) to the output directory
+    - create repodata.json
+
+
+    pkg_map
+        package map created from build_packages
+
+    output_dir
+        output directory to install packages into
+    """
+
+    output_dir.mkdir(exist_ok=True, parents=True)
     copy_packages_to_dist_dir(pkg_map.values(), output_dir)
     package_data = generate_repodata(output_dir, pkg_map)
 
@@ -714,10 +708,9 @@ def make_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
     return parser
 
 
-def main(args: argparse.Namespace) -> None:
-    packages_dir = Path(args.dir[0]).resolve()
-    outputdir = Path(args.output[0]).resolve()
-    outputdir.mkdir(exist_ok=True)
+def set_default_args(args: argparse.Namespace) -> argparse.Namespace:
+    args = copy.deepcopy(args)
+
     if args.cflags is None:
         args.cflags = common.get_make_flag("SIDE_MODULE_CFLAGS")
     if args.cxxflags is None:
@@ -728,7 +721,16 @@ def main(args: argparse.Namespace) -> None:
         args.target_install_dir = common.get_make_flag("TARGETINSTALLDIR")
     if args.host_install_dir is None:
         args.host_install_dir = common.get_make_flag("HOSTINSTALLDIR")
-    build_packages(packages_dir, outputdir, args)
+
+    return args
+
+
+def main(args: argparse.Namespace) -> None:
+    packages_dir = Path(args.dir[0]).resolve()
+    outputdir = Path(args.output[0]).resolve()
+    args = set_default_args(args)
+    pkg_map = build_packages(packages_dir, args)
+    install_packages(pkg_map, outputdir)
 
 
 if __name__ == "__main__":
