@@ -1,22 +1,91 @@
 import contextlib
 import functools
 import os
+import re
+import shutil
 import subprocess
 import sys
+import textwrap
+import zipfile
+from collections import deque
 from collections.abc import Generator, Iterable, Iterator, Mapping
 from pathlib import Path
+from typing import NoReturn
 
 import tomli
 from packaging.tags import Tag, compatible_tags, cpython_tags
 from packaging.utils import parse_wheel_filename
 
-from .io import parse_package_config
+from .io import MetaConfig
+
+BUILD_VARS: set[str] = {
+    "PATH",
+    "PYTHONPATH",
+    "PYODIDE_ROOT",
+    "PYTHONINCLUDE",
+    "NUMPY_LIB",
+    "PYODIDE_PACKAGE_ABI",
+    "HOME",
+    "HOSTINSTALLDIR",
+    "TARGETINSTALLDIR",
+    "SYSCONFIG_NAME",
+    "HOSTSITEPACKAGES",
+    "PYVERSION",
+    "PYMAJOR",
+    "PYMINOR",
+    "PYMICRO",
+    "CPYTHONBUILD",
+    "CPYTHONLIB",
+    "SIDE_MODULE_CFLAGS",
+    "SIDE_MODULE_CXXFLAGS",
+    "SIDE_MODULE_LDFLAGS",
+    "STDLIB_MODULE_CFLAGS",
+    "UNISOLATED_PACKAGES",
+    "WASM_LIBRARY_DIR",
+    "WASM_PKG_CONFIG_PATH",
+    "CARGO_BUILD_TARGET",
+    "CARGO_TARGET_WASM32_UNKNOWN_EMSCRIPTEN_LINKER",
+    "CARGO_HOME",
+    "RUSTFLAGS",
+    "PYODIDE_EMSCRIPTEN_VERSION",
+    "PLATFORM_TRIPLET",
+    "SYSCONFIGDATA_DIR",
+}
 
 
 def emscripten_version() -> str:
     return get_make_flag("PYODIDE_EMSCRIPTEN_VERSION")
 
 
+def get_emscripten_version_info() -> str:
+    """Extracted for testing purposes."""
+    return subprocess.run(["emcc", "-v"], capture_output=True, encoding="utf8").stderr
+
+
+def check_emscripten_version() -> None:
+    needed_version = emscripten_version()
+    try:
+        version_info = get_emscripten_version_info()
+    except FileNotFoundError:
+        raise RuntimeError(
+            f"No Emscripten compiler found. Need Emscripten version {needed_version}"
+        ) from None
+    installed_version = None
+    try:
+        for x in reversed(version_info.partition("\n")[0].split(" ")):
+            if re.match(r"[0-9]+\.[0-9]+\.[0-9]+", x):
+                installed_version = x
+                break
+    except Exception:
+        raise RuntimeError("Failed to determine Emscripten version.") from None
+    if installed_version is None:
+        raise RuntimeError("Failed to determine Emscripten version.")
+    if installed_version != needed_version:
+        raise RuntimeError(
+            f"Incorrect Emscripten version {installed_version}. Need Emscripten version {needed_version}"
+        )
+
+
 def platform() -> str:
     emscripten_version = get_make_flag("PYODIDE_EMSCRIPTEN_VERSION")
     version = emscripten_version.replace(".", "_")
@@ -61,12 +130,60 @@ def find_matching_wheels(wheel_paths: Iterable[Path]) -> Iterator[Path]:
                 yield wheel_path
 
 
-UNVENDORED_STDLIB_MODULES = {"test", "distutils"}
+def parse_top_level_import_name(whlfile: Path) -> list[str] | None:
+    """
+    Parse the top-level import names from a wheel file.
+    """
+
+    if not whlfile.name.endswith(".whl"):
+        raise RuntimeError(f"{whlfile} is not a wheel file.")
+
+    whlzip = zipfile.Path(whlfile)
+
+    def _valid_package_name(dirname: str) -> bool:
+        return all([invalid_chr not in dirname for invalid_chr in ".- "])
+
+    def _has_python_file(subdir: zipfile.Path) -> bool:
+        queue = deque([subdir])
+        while queue:
+            nested_subdir = queue.pop()
+            for subfile in nested_subdir.iterdir():
+                if subfile.is_file() and subfile.name.endswith(".py"):
+                    return True
+                elif subfile.is_dir() and _valid_package_name(subfile.name):
+                    queue.append(subfile)
+
+        return False
+
+    # If there is no top_level.txt file, we will find top level imports by
+    # 1) a python file on a top-level directory
+    # 2) a sub directory with __init__.py
+    # following: https://github.com/pypa/setuptools/blob/d680efc8b4cd9aa388d07d3e298b870d26e9e04b/setuptools/discovery.py#L122
+    top_level_imports = []
+    for subdir in whlzip.iterdir():
+        if subdir.is_file() and subdir.name.endswith(".py"):
+            top_level_imports.append(subdir.name[:-3])
+        elif subdir.is_dir() and _valid_package_name(subdir.name):
+            if _has_python_file(subdir):
+                top_level_imports.append(subdir.name)
+
+    if not top_level_imports:
+        print(f"Warning: failed to parse top level import name from {whlfile}.")
+        return None
+
+    return top_level_imports
+
 
 ALWAYS_PACKAGES = {
     "pyparsing",
     "packaging",
     "micropip",
+    "distutils",
+    "test",
+    "ssl",
+    "lzma",
+    "sqlite3",
+    "hashlib",
 }
 
 CORE_PACKAGES = {
@@ -79,7 +196,6 @@ CORE_PACKAGES = {
     "fpcast-test",
     "sharedlib-test-py",
     "cpp-exceptions-test",
-    "ssl",
     "pytest",
     "tblib",
 }
@@ -123,7 +239,6 @@ def _parse_package_subset(query: str | None) -> set[str]:
 
     packages = {el.strip() for el in query.split(",")}
     packages.update(ALWAYS_PACKAGES)
-    packages.update(UNVENDORED_STDLIB_MODULES)
     # handle meta-packages
     if "core" in packages:
         packages |= CORE_PACKAGES
@@ -132,10 +247,6 @@ def _parse_package_subset(query: str | None) -> set[str]:
         packages |= CORE_PACKAGES | CORE_SCIPY_PACKAGES
         packages.discard("min-scipy-stack")
 
-    # Hack to deal with the circular dependence between soupsieve and
-    # beautifulsoup4
-    if "beautifulsoup4" in packages:
-        packages.add("soupsieve")
     packages.discard("")
     return packages
 
@@ -147,7 +258,6 @@ def get_make_flag(name: str) -> str:
         SIDE_MODULE_LDFLAGS
         SIDE_MODULE_CFLAGS
         SIDE_MODULE_CXXFLAGS
-        TOOLSDIR
     """
     return get_make_environment_vars()[name]
 
@@ -179,6 +289,10 @@ def get_make_environment_vars() -> dict[str, str]:
         equalPos = line.find("=")
         if equalPos != -1:
             varname = line[0:equalPos]
+
+            if varname not in BUILD_VARS:
+                continue
+
             value = line[equalPos + 1 :]
             value = value.strip("'").strip()
             environment[varname] = value
@@ -204,8 +318,8 @@ def search_pyodide_root(curdir: str | Path, *, max_depth: int = 5) -> Path:
         try:
             with pyproject_file.open("rb") as f:
                 configs = tomli.load(f)
-        except tomli.TOMLDecodeError:
-            raise ValueError(f"Could not parse {pyproject_file}.")
+        except tomli.TOMLDecodeError as e:
+            raise ValueError(f"Could not parse {pyproject_file}.") from e
 
         if "tool" in configs and "pyodide" in configs["tool"]:
             return base
@@ -260,12 +374,13 @@ def get_unisolated_packages() -> list[str]:
         unisolated_packages = unisolated_file.read_text().splitlines()
     else:
         unisolated_packages = []
-        for pkg in (PYODIDE_ROOT / "packages").glob("**/meta.yaml"):
-            config = parse_package_config(pkg, check=False)
-            if config.get("build", {}).get("cross-build-env", False):
-                unisolated_packages.append(config["package"]["name"])
-        # TODO: remove setuptools_rust from this when they release the next version.
-        unisolated_packages.append("setuptools_rust")
+        for pkg in (PYODIDE_ROOT / "packages").glob("*/meta.yaml"):
+            try:
+                config = MetaConfig.from_yaml(pkg)
+            except Exception as e:
+                raise ValueError(f"Could not parse {pkg}.") from e
+            if config.build.cross_build_env:
+                unisolated_packages.append(config.package.name)
     os.environ["UNISOLATED_PACKAGES"] = json.dumps(unisolated_packages)
     return unisolated_packages
 
@@ -280,3 +395,22 @@ def replace_env(build_env: Mapping[str, str]) -> Generator[None, None, None]:
     finally:
         os.environ.clear()
         os.environ.update(old_environ)
+
+
+def exit_with_stdio(result: subprocess.CompletedProcess[str]) -> NoReturn:
+    if result.stdout:
+        print("  stdout:")
+        print(textwrap.indent(result.stdout, "    "))
+    if result.stderr:
+        print("  stderr:")
+        print(textwrap.indent(result.stderr, "    "))
+    raise SystemExit(result.returncode)
+
+
+def in_xbuildenv() -> bool:
+    pyodide_root = get_pyodide_root()
+    return pyodide_root.name == "pyodide-root"
+
+
+def find_missing_executables(executables: list[str]) -> list[str]:
+    return list(filter(lambda exe: shutil.which(exe) is None, executables))
