@@ -1,19 +1,26 @@
 #!/usr/bin/env python3
 
 import argparse
+import contextlib
 import json
 import os
 import shutil
 import subprocess
 import sys
+import tempfile
 import urllib.error
 import urllib.request
 import warnings
+from collections.abc import Iterator
 from pathlib import Path
 from typing import Any, Literal, NoReturn, TypedDict
+from urllib import request
 
+from packaging.version import Version
 from ruamel.yaml import YAML
 
+from .common import parse_top_level_import_name
+
 
 class URLDict(TypedDict):
     comment_text: str
@@ -63,12 +70,14 @@ def _find_sdist(pypi_metadata: MetadataDict) -> URLDict | None:
     return None
 
 
-def _find_wheel(pypi_metadata: MetadataDict) -> URLDict | None:
+def _find_wheel(pypi_metadata: MetadataDict, native: bool = False) -> URLDict | None:
     """Get wheel file path from the metadata"""
+    predicate = lambda filename: filename.endswith(
+        ".whl" if native else "py3-none-any.whl"
+    )
+
     for entry in pypi_metadata["urls"]:
-        if entry["packagetype"] == "bdist_wheel" and entry["filename"].endswith(
-            "py3-none-any.whl"
-        ):
+        if entry["packagetype"] == "bdist_wheel" and predicate(entry["filename"]):
             return entry
     return None
 
@@ -110,11 +119,22 @@ def _get_metadata(package: str, version: str | None = None) -> MetadataDict:
         raise MkpkgFailedException(
             f"Failed to load metadata for {package}{version} from "
             f"https://pypi.org/pypi/{package}{version}/json: {e}"
-        )
+        ) from e
 
     return pypi_metadata
 
 
+@contextlib.contextmanager
+def _download_wheel(pypi_metadata: URLDict) -> Iterator[Path]:
+    response = request.urlopen(pypi_metadata["url"])
+    whlname = Path(response.geturl()).name
+
+    with tempfile.TemporaryDirectory() as tmpdirname:
+        whlpath = Path(tmpdirname, whlname)
+        whlpath.write_bytes(response.read())
+        yield whlpath
+
+
 def run_prettier(meta_path: str | Path) -> None:
     subprocess.run(["npx", "prettier", "-w", meta_path])
 
@@ -142,6 +162,13 @@ def make_package(
         sources = ["wheel", "sdist"]
     dist_metadata = _find_dist(pypi_metadata, sources)
 
+    native_wheel_metadata = _find_wheel(pypi_metadata, native=True)
+
+    top_level = None
+    if native_wheel_metadata is not None:
+        with _download_wheel(native_wheel_metadata) as native_wheel_path:
+            top_level = parse_top_level_import_name(native_wheel_path)
+
     url = dist_metadata["url"]
     sha256 = dist_metadata["digests"]["sha256"]
     version = pypi_metadata["info"]["version"]
@@ -152,9 +179,12 @@ def make_package(
     pypi = "https://pypi.org/project/" + package
 
     yaml_content = {
-        "package": {"name": package, "version": version},
+        "package": {
+            "name": package,
+            "version": version,
+            "top-level": top_level or ["PUT_TOP_LEVEL_IMPORT_NAMES_HERE"],
+        },
         "source": {"url": url, "sha256": sha256},
-        "test": {"imports": [package]},
         "about": {
             "home": homepage,
             "PyPI": pypi,
@@ -170,6 +200,7 @@ def make_package(
     if meta_path.exists():
         raise MkpkgFailedException(f"The package {package} already exists")
 
+    yaml.representer.ignore_aliases = lambda *_: True
     yaml.dump(yaml_content, meta_path)
     try:
         run_prettier(meta_path)
@@ -234,8 +265,8 @@ def update_package(
         old_fmt = "sdist"
 
     pypi_metadata = _get_metadata(package, version)
-    pypi_ver = pypi_metadata["info"]["version"]
-    local_ver = yaml_content["package"]["version"]
+    pypi_ver = Version(pypi_metadata["info"]["version"])
+    local_ver = Version(yaml_content["package"]["version"])
     already_up_to_date = pypi_ver <= local_ver and (
         source_fmt is None or source_fmt == old_fmt
     )
@@ -245,7 +276,7 @@ def update_package(
 
     print(f"{package} is out of date: {local_ver} <= {pypi_ver}.")
 
-    if "patches" in yaml_content["source"]:
+    if yaml_content["source"].get("patches"):
         if update_patched:
             warn(
                 f"Pyodide applies patches to {package}. Update the "
@@ -304,6 +335,12 @@ complex things.""".strip()
         help="Package version string, "
         "e.g. v1.2.1 (defaults to latest stable release)",
     )
+    parser.add_argument(
+        "--recipe-dir",
+        type=str,
+        default="packages",
+        help="Directory to create the recipe in (defaults to PYODIDE_ROOT/packages)",
+    )
     return parser
 
 
@@ -312,13 +349,16 @@ def main(args: argparse.Namespace) -> None:
     if PYODIDE_ROOT is None:
         raise ValueError("PYODIDE_ROOT is not set")
 
-    PACKAGES_ROOT = Path(PYODIDE_ROOT) / "packages"
+    if os.path.isabs(args.recipe_dir):
+        recipe_dir = Path(args.recipe_dir)
+    else:
+        recipe_dir = (Path(PYODIDE_ROOT) / args.recipe_dir).resolve()
 
     try:
         package = args.package[0]
         if args.update:
             update_package(
-                PACKAGES_ROOT,
+                recipe_dir,
                 package,
                 args.version,
                 update_patched=True,
@@ -327,16 +367,14 @@ def main(args: argparse.Namespace) -> None:
             return
         if args.update_if_not_patched:
             update_package(
-                PACKAGES_ROOT,
+                recipe_dir,
                 package,
                 args.version,
                 update_patched=False,
                 source_fmt=args.source_format,
             )
             return
-        make_package(
-            PACKAGES_ROOT, package, args.version, source_fmt=args.source_format
-        )
+        make_package(recipe_dir, package, args.version, source_fmt=args.source_format)
     except MkpkgFailedException as e:
         # This produces two types of error messages:
         #
